#include "Message.h"

#define MIN(a,b) ( ((a)<(b)) ? (a) : (b) );

int counter = 0;

void DumpMessage(Message_t *msg)
{
	char textbuf[10000];
	textbuf[0]  = '\0';
	char *buf = (char*)msg;
	for (int i=0; i<sizeof(*msg); i++) {
		sprintf(textbuf+strlen(textbuf), "%02hhx", buf[i]);
	}
	Log("%s", textbuf);
}

Message::Message(char type) {
	//memset(&buffer, '.', sizeof(buffer));
	buffer.header.type = type;
	//buffer.serial = counter++; // moved to Send()
	//Log("type=%c", type);
};

Message::Message(Message_t *msg) {
	memcpy(&buffer, msg, sizeof(*msg));
	//Log("message copy done");
};

int Message::Send(SOCKET s) { 
	buffer.header.serial = counter++;
	return send(s, (const char*) &buffer, GetSize(), 0); 
};

int Message::Send(SOCKET s, struct sockaddr *client) {
	buffer.header.serial = counter++;
	return sendto(s, (const char*) &buffer, GetSize(), 0, client, sizeof(*client));
};

void Message::Send(SOCKET s, vector<struct sockaddr> *clients)
{
	for (int i=0; i<clients->size(); i++) {
		Send(s, &(clients->at(i)));
		counter--; // increased in Send()
	}

	counter++; // only once per packet, not for each client!
}

int GetMessageSize(Message_t *buffer)
{
	switch ( buffer->header.type ) {
	case MESSAGE_TYPE_START:
		return sizeof(buffer->header) + sizeof(buffer->msg.start);
	case MESSAGE_TYPE_VESSEL:
		return sizeof(buffer->header) + sizeof(buffer->msg.vessel);
	case MESSAGE_TYPE_VESSEL2:
		return sizeof(buffer->header) + sizeof(buffer->msg.vessel2);
	case MESSAGE_TYPE_VESSELSTATE:
		return sizeof(buffer->header) + sizeof(buffer->msg.vesselstate.subheader) + buffer->msg.vesselstate.subheader.length;
	case MESSAGE_TYPE_KEYBOARD:
		return sizeof(buffer->header) + sizeof(buffer->msg.keyboard);
	case MESSAGE_TYPE_VESSELEXTRA:
		return sizeof(buffer->header) + sizeof(buffer->msg.vesselextra);
	case MESSAGE_TYPE_AIRFOIL:
		return sizeof(buffer->header) + sizeof(buffer->msg.airfoil);
		case MESSAGE_TYPE_ANIMATION:
		return sizeof(buffer->header) + sizeof(buffer->msg.animation);
	case MESSAGE_TYPE_XR:
		return sizeof(buffer->header) + sizeof(buffer->msg.xr);
	case MESSAGE_TYPE_FOCUS:
		return sizeof(buffer->header) + sizeof(buffer->msg.focus);
	case MESSAGE_TYPE_END:
	case MESSAGE_TYPE_ACK:
	case MESSAGE_TYPE_HELLO:
		return sizeof(buffer->header);
	}

	Log("Unhandled message type %d", buffer->header.type);

	return -1;  
};

int Message::GetSize(void)
{
	return GetMessageSize(&buffer);
}

VesselMessage::VesselMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_VESSEL)
{
	buffer.msg.vessel.mjd = oapiGetSimMJD();

	VESSEL *v = oapiGetVesselInterface(vessel);

	_ASSERTE(strlen(v->GetName()) < MSG_MAXSTRLEN);
	_ASSERTE(strlen(v->GetClassNameA()) < MSG_MAXSTRLEN);
	strcpy(buffer.msg.vessel.name, v->GetName());
	strcpy(buffer.msg.vessel.classname, v->GetClassNameA());

	v->GetStatus(buffer.msg.vessel.status);
	buffer.msg.vessel.rbody[0] = '\0';
	buffer.msg.vessel.base[0] = '\0';

	if ( buffer.msg.vessel.status.rbody )
		oapiGetObjectName(buffer.msg.vessel.status.rbody, buffer.msg.vessel.rbody, MSG_MAXSTRLEN);
	if ( buffer.msg.vessel.status.base )
		oapiGetObjectName(buffer.msg.vessel.status.base, buffer.msg.vessel.base, MSG_MAXSTRLEN);
}

VesselMessage::VesselMessage(Message_t *msg) : Message(msg)
{
	if ( buffer.msg.vessel.status.rbody )
		buffer.msg.vessel.status.rbody = oapiGetObjectByName(msg->msg.vessel.rbody);

	if ( buffer.msg.vessel.status.base )
		buffer.msg.vessel.status.base = oapiGetObjectByName(msg->msg.vessel.base);

	// Fix jitter
	double hisMjd = buffer.msg.vessel.mjd;
	double myMjd = oapiGetSimMJD();	
	double dt = (myMjd-hisMjd)*86400;

	VESSELSTATUS *hisStatus = GetStatus();
	hisStatus->rpos.x = hisStatus->rpos.x + hisStatus->rvel.x*dt;
	hisStatus->rpos.y = hisStatus->rpos.y + hisStatus->rvel.y*dt;
	hisStatus->rpos.z = hisStatus->rpos.z + hisStatus->rvel.z*dt;
}

void VesselMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	//vmsg->GetStatus()->status = 0; // enforce freeflight mode

	if ( hVessel ) {
		// vessel already exists, reconfigure
		VESSEL *v = oapiGetVesselInterface(hVessel);
		v->DefSetState( GetStatus() );
	} else {
		// create from scratch
		oapiCreateVessel( GetName(), GetClassNameA(), *(GetStatus()) );
	}
}

Vessel2Message::Vessel2Message(OBJHANDLE vessel) : Message(MESSAGE_TYPE_VESSEL2)
{
	buffer.msg.vessel2.mjd = oapiGetSimMJD();

	VESSEL *v = oapiGetVesselInterface(vessel);
	_ASSERTE(strlen(v->GetName()) < MSG_MAXSTRLEN-1);
	_ASSERTE(strlen(v->GetClassNameA()) < MSG_MAXSTRLEN-1);

	strcpy(buffer.msg.vessel2.name, v->GetName());
	strcpy(buffer.msg.vessel2.classname, v->GetClassNameA());
	//Log("Vessel2Message: name=%s, class=%s", v->GetName(), v->GetClassNameA());

	buffer.msg.vessel2.status.version = 2;
	buffer.msg.vessel2.status.flag = VS_FUELLIST | VS_THRUSTLIST | VS_DOCKINFOLIST;
	buffer.msg.vessel2.status.fuel = buffer.msg.vessel2.fuelspec;
	buffer.msg.vessel2.status.thruster = buffer.msg.vessel2.thrustspec;
	buffer.msg.vessel2.status.dockinfo = buffer.msg.vessel2.dockinfo;
	v->GetStatusEx(&buffer.msg.vessel2.status);

	buffer.msg.vessel2.status.nfuel = MIN(buffer.msg.vessel2.status.nfuel, MAX_FUEL_TANKS);
	buffer.msg.vessel2.status.nthruster = MIN(buffer.msg.vessel2.status.nthruster, MAX_THRUSTERS);
	buffer.msg.vessel2.status.ndockinfo = MIN(buffer.msg.vessel2.status.ndockinfo, MAX_DOCKING_PORTS);

	buffer.msg.vessel2.rbody[0] = '\0';
	buffer.msg.vessel2.base[0] = '\0';

	if ( buffer.msg.vessel2.status.rbody )
		oapiGetObjectName(buffer.msg.vessel2.status.rbody, buffer.msg.vessel2.rbody, MSG_MAXSTRLEN);
	if ( buffer.msg.vessel2.status.base )
		oapiGetObjectName(buffer.msg.vessel2.status.base, buffer.msg.vessel2.base, MSG_MAXSTRLEN);

	for ( int i=0; i<buffer.msg.vessel2.status.ndockinfo; i++ ) {
		if ( buffer.msg.vessel2.dockinfo[i].rvessel ) {
			VESSEL *dockedVessel = oapiGetVesselInterface(buffer.msg.vessel2.dockinfo[i].rvessel);
			strcpy(buffer.msg.vessel2.dockedVessel[i], dockedVessel->GetName());
		}
	}

}

Vessel2Message::Vessel2Message(Message_t *msg) : Message(msg)
{
	if ( buffer.msg.vessel2.status.rbody )
		buffer.msg.vessel2.status.rbody = oapiGetObjectByName(msg->msg.vessel2.rbody);

	if ( buffer.msg.vessel2.status.base )
		buffer.msg.vessel2.status.base = oapiGetObjectByName(msg->msg.vessel2.base);

	buffer.msg.vessel2.status.fuel = buffer.msg.vessel2.fuelspec;
	buffer.msg.vessel2.status.thruster = buffer.msg.vessel2.thrustspec;
	buffer.msg.vessel2.status.dockinfo = buffer.msg.vessel2.dockinfo;

	for ( int i=0; i<buffer.msg.vessel2.status.ndockinfo; i++ ) {
		if ( buffer.msg.vessel2.dockinfo[i].rvessel ) {
			buffer.msg.vessel2.dockinfo[i].rvessel = oapiGetVesselByName(buffer.msg.vessel2.dockedVessel[i]);
		}
	}

	// Fix jitter
	double hisMjd = buffer.msg.vessel2.mjd;
	double myMjd = oapiGetSimMJD();	
	double dt = (myMjd-hisMjd)*86400;

	VESSELSTATUS2 *hisStatus = GetStatus();
	hisStatus->rpos.x = hisStatus->rpos.x + hisStatus->rvel.x*dt;
	hisStatus->rpos.y = hisStatus->rpos.y + hisStatus->rvel.y*dt;
	hisStatus->rpos.z = hisStatus->rpos.z + hisStatus->rvel.z*dt;
}

void Vessel2Message::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	//vmsg->GetStatus()->status = 0; // enforce freeflight mode

	if ( hVessel ) {
		// vessel already exists, reconfigure
		VESSEL *v = oapiGetVesselInterface(hVessel);
		v->DefSetStateEx( GetStatus() );
	} else {
		// create from scratch
		oapiCreateVesselEx( GetName(), GetClassNameA(), GetStatus() );
	}
}



VesselStateMessage::VesselStateMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_VESSELSTATE)
{
	buffer.msg.vesselstate.subheader.mjd = oapiGetSimMJD();
	VESSEL *v = oapiGetVesselInterface(vessel);
	strcpy(buffer.msg.vesselstate.subheader.name, v->GetName());

	VESSEL2 *v2 = (VESSEL2*) v;

	FILEHANDLE scn = oapiOpenFile("MASTERSLAVE_m", FILE_OUT, ROOT);
	((VESSEL2*)v)->clbkSaveState(scn);
	oapiCloseFile(scn, FILE_OUT);

	buffer.msg.vesselstate.buf[0] = '\0';
	scn = oapiOpenFile("MASTERSLAVE_m", FILE_IN, ROOT);
	char *line;
	int pos = 0;
	while (oapiReadScenario_nextline (scn, line)) {
		int len = strlen(line);
		strcpy(buffer.msg.vesselstate.buf+pos, line);
		buffer.msg.vesselstate.buf[pos+len] = '\n';
		pos = pos + len + 1;
	}
	buffer.msg.vesselstate.buf[pos] = '\0';

	buffer.msg.vesselstate.subheader.length = pos - 1;
	oapiCloseFile(scn, FILE_IN);
}

void VesselStateMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	if ( ! hVessel )
		return ;
	
	VESSEL *v = oapiGetVesselInterface(hVessel);
	if ( v->Version() < 2 )
		return ;

	//Log("Text-based sync for %s", v->GetName());

	char curDir[1024];
	GetCurrentDirectory(1024, curDir);

	char file[1024];
	sprintf(file, "%s\\MASTERSLAVE_%d", curDir, GetCurrentProcessId());

	//Log("STATE: %d bytes", buffer.msg.vesselstate.subheader.length);
	buffer.msg.vesselstate.buf[buffer.msg.vesselstate.subheader.length] = '\0';
	//Log("STATE: %s", buffer.msg.vesselstate.buf);

	FILE *f = fopen(file, "w");
	fwrite(buffer.msg.vesselstate.buf, 1, buffer.msg.vesselstate.subheader.length, f);
	fclose(f);

	VESSELSTATUS2 status;
	v->GetStatusEx(&status);

	char scnFile[1024];
	sprintf(scnFile, "MASTERSLAVE_%d", GetCurrentProcessId());
	FILEHANDLE scn = oapiOpenFile(scnFile, FILE_IN, ROOT);
	((VESSEL2*)v)->clbkLoadStateEx(scn, &status);
	oapiCloseFile(scn, FILE_IN);

	DeleteFile(file);

	// Fix jitter
	double hisMjd = buffer.msg.vesselstate.subheader.mjd;
	double myMjd = oapiGetSimMJD();	
	double dt = (myMjd-hisMjd)*86400;

	v->GetStatusEx(&status);
	status.rpos.x = status.rpos.x + status.rvel.x*dt;
	status.rpos.y = status.rpos.y + status.rvel.y*dt;
	status.rpos.z = status.rpos.z + status.rvel.z*dt;
	v->DefSetStateEx(&status);
}

AirfoilMessage::AirfoilMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_AIRFOIL)
{
	VESSEL *v = oapiGetVesselInterface(vessel);
	strcpy(buffer.msg.airfoil.vesselName, v->GetName());
	buffer.msg.airfoil.elevator = v->GetControlSurfaceLevel(AIRCTRL_ELEVATOR);
	buffer.msg.airfoil.rudder = v->GetControlSurfaceLevel(AIRCTRL_RUDDER);
	buffer.msg.airfoil.aileron = v->GetControlSurfaceLevel(AIRCTRL_AILERON);
	buffer.msg.airfoil.flap = v->GetControlSurfaceLevel(AIRCTRL_FLAP);
	buffer.msg.airfoil.elevatorTrim = v->GetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM);
	buffer.msg.airfoil.rudderTrim = v->GetControlSurfaceLevel(AIRCTRL_RUDDERTRIM);
}

void AirfoilMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	if ( ! hVessel )
		return ;

	VESSEL *v = oapiGetVesselInterface(hVessel);
	v->SetControlSurfaceLevel(AIRCTRL_ELEVATOR, buffer.msg.airfoil.elevator, TRUE);
	v->SetControlSurfaceLevel(AIRCTRL_RUDDER, buffer.msg.airfoil.rudder, TRUE);
	v->SetControlSurfaceLevel(AIRCTRL_AILERON, buffer.msg.airfoil.aileron, TRUE);
	v->SetControlSurfaceLevel(AIRCTRL_FLAP, buffer.msg.airfoil.flap, TRUE);
	v->SetControlSurfaceLevel(AIRCTRL_ELEVATORTRIM, buffer.msg.airfoil.elevatorTrim, TRUE);
	v->SetControlSurfaceLevel(AIRCTRL_RUDDERTRIM, buffer.msg.airfoil.rudderTrim, TRUE);
}

AnimationMessage::AnimationMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_ANIMATION)
{
	VESSEL *v = oapiGetVesselInterface(vessel);
	strcpy(buffer.msg.animation.vesselName, v->GetName());
	ANIMATION *anim;
	buffer.msg.animation.nAnimations = v->GetAnimPtr(&anim);
	buffer.msg.animation.nAnimations = MIN(buffer.msg.animation.nAnimations, MAX_ANIMATIONS);

	for (int i=0; i<buffer.msg.animation.nAnimations; i++) {
		buffer.msg.animation.state[i] = anim[i].state;
	}
}

void AnimationMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	if ( ! hVessel )
		return ;

	VESSEL *v = oapiGetVesselInterface(hVessel);
	for (int i=0; i<buffer.msg.animation.nAnimations; i++) {
		//Log("AnimationMessage::UpdateVessel %s %i %f", GetName(), i, buffer.msg.animation.state[i]);
		v->SetAnimation(i, buffer.msg.animation.state[i]);
	}
}

VesselExtraMessage::VesselExtraMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_VESSELEXTRA)
{
	VESSEL *v = oapiGetVesselInterface(vessel);
	strcpy(buffer.msg.vesselextra.vesselName, v->GetName());
	
	buffer.msg.vesselextra.rcsMode = v->GetAttitudeMode();

	buffer.msg.vesselextra.navModeState = 0;
	// See Orbiter SDK help on navmode identifiers
	for (int i=1; i<=7; i++) {
		if ( v->GetNavmodeState(i) )
			buffer.msg.vesselextra.navModeState |= 1 << i;
	}

	buffer.msg.vesselextra.adCtrlMode = v->GetADCtrlMode();

	int nNav = MIN(v->GetNavCount(), MAX_NAV_RADIOS);
	for (int i=0; i<nNav; i++)
		buffer.msg.vesselextra.navChannel[i] = (WORD) v->GetNavChannel(i);
}

void VesselExtraMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	if ( ! hVessel )
		return ;

	VESSEL *v = oapiGetVesselInterface(hVessel);
	v->SetAttitudeMode(buffer.msg.vesselextra.rcsMode);

	// See Orbiter SDK help on navmode identifiers
	for (int i=1; i<=7; i++) {
		int mode = (buffer.msg.vesselextra.navModeState & ( 1 << i )) >> i;
		if ( mode ) 
			v->ActivateNavmode(i);
		else
			v->DeactivateNavmode(i);
	}

	v->SetADCtrlMode(buffer.msg.vesselextra.adCtrlMode);

	for (int i=0; i<v->GetNavCount(); i++) {
		v->SetNavChannel(i, buffer.msg.vesselextra.navChannel[i]);
	}
}

Message *CreateVesselMessage(OBJHANDLE vessel)
{
	VESSEL *v = oapiGetVesselInterface(vessel);
	
	if ( v->Version() < 2 )
		return (Message*) new VesselMessage(vessel);

	if ( v->Version() == 2 )
		return (Message*) new Vessel2Message(vessel);

	if ( v->Version() == 3 )
		return (Message*) new Vessel2Message(vessel);

	Log("vessel %s: unsupported vessel interface version %d", v->GetName(), v->Version());
	return NULL;
}

Message *CreateVesselStateMessage(OBJHANDLE vessel)
{
	VESSEL *v = oapiGetVesselInterface(vessel);
	if ( v->Version() < 2 )
		return NULL;

	return (Message*) new VesselStateMessage(vessel);
}

XRMessage::XRMessage(OBJHANDLE vessel) : Message(MESSAGE_TYPE_XR)
{
	bValid = FALSE;
	VESSEL *v = oapiGetVesselInterface(vessel);	
	strcpy(buffer.msg.xr.vesselName, v->GetName());

	if ( ! XRVesselCtrl::IsXRVesselCtrl(v) )
		return ;

	XRVesselCtrl *xr = (XRVesselCtrl *)v;

	XRSystemStatusRead xrStatus;
	xr->GetXRSystemStatus(xrStatus);
	memcpy(&buffer.msg.xr.xrSystemStatus, &xrStatus, sizeof(XRSystemStatusWrite));
	
	for (int i=0; i<XR_LIGHTS; i++) {
		buffer.msg.xr.xrLightsStatus[i] = xr->GetExteriorLight((XRLight) i);
	}
	
	for (int i=0; i<XR_DOORS; i++) {
		buffer.msg.xr.xrDoorStatus[i] = xr->GetDoorState((XRDoorID) i);
	}

	XREngineStateRead engineStatus;
	for (int i=0; i<XR_ENGINES; i++) {
		xr->GetEngineState((XREngineID) i, engineStatus);
		memcpy(&buffer.msg.xr.xrEngineStatus[i], &engineStatus, sizeof(XREngineStateWrite));
	}

	buffer.msg.xr.xrCenterOfGravity = xr->GetCenterOfGravity();
	buffer.msg.xr.xrRCSDockingMode = xr->IsRCSDockingMode();
	buffer.msg.xr.xrElevatorEVAPortActive = xr->IsElevatorEVAPortActive();
	bValid = TRUE;
}

void XRMessage::UpdateVessel(void)
{
	OBJHANDLE hVessel = oapiGetVesselByName(GetName());
	if ( ! hVessel )
		return ;

	VESSEL *v = oapiGetVesselInterface(hVessel);	

	if ( ! XRVesselCtrl::IsXRVesselCtrl(v) )
		return ;

	XRVesselCtrl *xr = (XRVesselCtrl *)v;
	xr->SetXRSystemStatus(buffer.msg.xr.xrSystemStatus);
	
	for (int i=0; i<XR_LIGHTS; i++) {
		xr->SetExteriorLight((XRLight)i, buffer.msg.xr.xrLightsStatus[i]);
	}

	for (int i=0; i<XR_DOORS; i++) {
		if ( i != 12 ) // This is not APU
			if ( buffer.msg.xr.xrDoorStatus[12] != XRDS_Open ) // and APU is off
				continue;

		xr->SetDoorState((XRDoorID)i, (XRDoorState) buffer.msg.xr.xrDoorStatus[i]);
	}

	for (int i=0; i<XR_ENGINES; i++) {
		if ( ((XREngineID)i == XRE_HoverFore) || ((XREngineID)i == XRE_HoverAft) ) {
			if ( xr->GetDoorState(XRD_HoverDoors) != XRDS_Open ) 
				continue;
		}
		if ( ((XREngineID)i == XRE_ScramLeft) || ((XREngineID)i == XRE_ScramRight) ) { 
			if ( xr->GetDoorState(XRD_ScramDoors) != XRDS_Open ) 
				continue;
		}
		if ( ((XREngineID)i == XRE_RetroLeft) || ((XREngineID)i == XRE_RetroRight) ) { 
			if ( xr->GetDoorState(XRD_RetroDoors) != XRDS_Open ) 
				continue;
		}

		xr->SetEngineState((XREngineID) i, buffer.msg.xr.xrEngineStatus[i]);
	}

	if ( buffer.msg.xr.xrDoorStatus[12] == XRDS_Open ) // APU is off
		xr->ShiftCenterOfGravity( buffer.msg.xr.xrCenterOfGravity - xr->GetCenterOfGravity() );

	xr->SetRCSDockingMode( buffer.msg.xr.xrRCSDockingMode );
	xr->SetElevatorEVAPortActive( buffer.msg.xr.xrElevatorEVAPortActive );
}

FocusMessage::FocusMessage(void) : Message(MESSAGE_TYPE_FOCUS)
{
	VESSEL *v = oapiGetFocusInterface();
	strcpy(buffer.msg.focus.vesselName, v->GetName());
}

void FocusMessage::SwitchFocus(void)
{
	HANDLE hVessel = oapiGetVesselByName(buffer.msg.focus.vesselName);
	if ( ! hVessel ) 
		return ;

	oapiSetFocusObject(hVessel);
}

KeyboardMessage KeyboardMessageBuf;
HHOOK hook;

void KeyboardMessage::HookProc(int ncode, WPARAM wparam, LPARAM lparam)
{
	if(ncode>=0) {
		//Log("Keyboard (hook): ncode=%d wparam=0x%04x lparam=0x%04x", ncode, wparam, lparam);

		if ( ( buffer.msg.keyboard.events[buffer.msg.keyboard.len].ncode == ncode ) && 
			 ( buffer.msg.keyboard.events[buffer.msg.keyboard.len].wparam == wparam ) &&
			 ( buffer.msg.keyboard.events[buffer.msg.keyboard.len].lparam == lparam ) )
			 return ; // discard duplicates

		if ( buffer.msg.keyboard.len >= MAX_KEYBOARD_EVENTS ) {
			//Log("Keyboard buffer overflow, event discarded");
			return;
		}

		buffer.msg.keyboard.events[buffer.msg.keyboard.len].ncode = ncode;
		buffer.msg.keyboard.events[buffer.msg.keyboard.len].lparam = lparam;
		buffer.msg.keyboard.events[buffer.msg.keyboard.len].wparam = wparam;
		buffer.msg.keyboard.len = buffer.msg.keyboard.len + 1;
	}
}

DLLCLBK LRESULT CALLBACK hookproc(int ncode, WPARAM wparam, LPARAM lparam)
{
	KeyboardMessageBuf.HookProc(ncode, wparam, lparam);

	return CallNextHookEx(hook,ncode,wparam,lparam);
}


void HookKeyboard(char *moduleName)
{
	hook = SetWindowsHookEx( WH_KEYBOARD, hookproc, GetModuleHandle(moduleName), NULL);
	if( ! hook )
		Log("Unable to install keyboard hook");
}
 
void UnHookKeyboard(void)
{
	UnhookWindowsHookEx(hook);
}

KeyboardMessage::KeyboardMessage(void) : Message(MESSAGE_TYPE_KEYBOARD)
{
	buffer.msg.keyboard.len = 0;
}

char kstate[256] = {0};

void KeyboardMessage::SetState(void)
{
	for (int i=0; i<buffer.msg.keyboard.len; i++) {
		keyevent_t *ev = &buffer.msg.keyboard.events[i];

		WORD code = (ev->lparam & 0x00FF0000) >> 16;
		//Log("Keyboard (net): ncode=%d wparam=%04x lparam=%04x code=%04x", ev->ncode, ev->wparam, ev->lparam, code); 

		VESSEL *v = oapiGetFocusInterface();
		if (((ev->lparam >> 24) & 0xf0) == 0xc0 ) {
			kstate[code] = 0;
			
			v->SendBufferedKey( code, false, kstate );
		} else {
			kstate[code] = 0x80;
			v->SendBufferedKey( code, true, kstate );
		}
	}
}

void KeyboardMessage::Clear(void)
{
	buffer.msg.keyboard.len = 0;
}

Message *ParseMessage(Message_t *msg)
{
	switch ( msg->header.type ) {
	case  MESSAGE_TYPE_START: return (Message*) new StartMessage(msg);
	case  MESSAGE_TYPE_END: return (Message*) new EndMessage(msg);
	case  MESSAGE_TYPE_VESSEL: return (Message*) new VesselMessage(msg);
	case  MESSAGE_TYPE_VESSEL2: return (Message*) new Vessel2Message(msg);
	case  MESSAGE_TYPE_VESSELSTATE: return (Message*) new VesselStateMessage(msg);
	case  MESSAGE_TYPE_KEYBOARD: return (Message*) new KeyboardMessage(msg);
	case  MESSAGE_TYPE_AIRFOIL: return (Message*) new AirfoilMessage(msg);
	case  MESSAGE_TYPE_ANIMATION: return (Message*) new AnimationMessage(msg);
	case  MESSAGE_TYPE_VESSELEXTRA: return (Message*) new VesselExtraMessage(msg);
	case  MESSAGE_TYPE_XR: return (Message*) new XRMessage(msg);
	case  MESSAGE_TYPE_FOCUS: return (Message*) new FocusMessage(msg);
	case  MESSAGE_TYPE_ACK: return (Message*) new AckMessage(msg);
	case  MESSAGE_TYPE_HELLO: return (Message*) new HelloMessage(msg);
	}

	Log("Unhandled message type %d", msg->header.type);

	return NULL;
}

Message *ReceiveMessage(SOCKET s, sockaddr *from, int *fromlen)
{
	Message_t msg;

	int len = recvfrom(s, (char*) &msg, sizeof(msg), 0 /* flags */, from, fromlen); 
	if ( len == SOCKET_ERROR ) {
		Log("recvfrom() error: %d", WSAGetLastError());
		return NULL;
	}

	if ( len != GetMessageSize(&msg) ) {
		return NULL;
	}

	//Log("Packet: serial=%d type=%d", msg.serial, msg.type);

	//DumpMessage(&msg);

	return ParseMessage(&msg);
}

Message *ReceiveMessage(SOCKET s)
{
	return ReceiveMessage(s, NULL, NULL);
}

Message *ReceiveMessage(SOCKET s, int timeOutMs)
{
	int cnt = 0;
	while ( ! MessageSocketHasData(s) ) {
		if ( ++cnt > timeOutMs )
			return MESSAGE_TIMED_OUT;

		Sleep(1);
	}

	return ReceiveMessage(s);
}