// NetDialog.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"
#include <string>

#define STRICT
#define ORBITER_MODULE
#include "Orbitersdk.h"

#include "Message.h"

HINSTANCE hDllInstance;

class OrbMaster : oapi::Module {
public:
	OrbMaster(HINSTANCE hDLL) : oapi::Module(hDLL) { bInitialized = FALSE;};
	virtual void clbkSimulationStart(RenderMode mode);
	virtual void clbkPostStep(double simt, double simdt, double mjd);
	virtual void clbkPreStep(double simt, double simdt, double mjd);
	virtual void clbkSimulationEnd(void);

private:
	bool bInitialized;
	SOCKET sock;
	double lastSaveT;
	int waitForClients;
	int waitForACK;
	int maxUpdateRate;
	double lastUpdateT;
	int readOnly;
	int debug;

	void Init(void);
	int FindClient(struct sockaddr_in *client_in);

	vector<struct sockaddr> clients;
	vector<double> clientLastAck;
	vector<char *>textSyncClasses;
};

void OrbMaster::Init(void)
{
	bInitialized = FALSE;

	FILEHANDLE f = oapiOpenFile("OrbMaster.cfg", FILE_IN, CONFIG);
	if ( ! oapiReadItem_int(f, "debug", debug ) ) {
		debug = 0;
	}
	if (debug) Log("Config: debug=%d", debug);

	if ( ! oapiReadItem_int(f, "waitForClients", waitForClients ) ) {
		waitForClients = 0;
	}
	if (debug) Log("Config: waitForClients=%d", waitForClients);

	if ( ! oapiReadItem_int(f, "waitForACK", waitForACK ) ) {
		waitForACK = 0;
	}
	if (debug) Log("Config: waitForACK=%d", waitForACK);

	if ( ! oapiReadItem_int(f, "maxUpdateRate", maxUpdateRate ) ) {
		maxUpdateRate = 30;
	}
	if (debug) Log("Config: maxUpdateRate=%d", maxUpdateRate);

	int serverPort;
	if ( ! oapiReadItem_int(f, "masterPort", serverPort ) ) {
		serverPort = 40000;
	}
	if (debug) Log("Config: masterPort=%d", serverPort);

	if ( ! oapiReadItem_int(f, "readOnly", readOnly ) ) {
		readOnly = 0;
	}
	if (debug) Log("Config: readOnly=%d", readOnly);

	char textSync[1024];
	if ( oapiReadItem_string(f, "textSyncClasses", textSync) ) {
		if (debug) Log("Config: textSyncClasses=%s", textSync);
		char *pch = strtok (textSync," ,");
		while (pch != NULL)
		{
			if (debug) Log("Config: using text sync for vessel class %s", pch);
			textSyncClasses.push_back(strdup(pch));
			pch = strtok (NULL, " ,.-");
		}
	}

	struct sockaddr_in serverAddr;
	int len = sizeof(serverAddr);
	
	WSADATA wsaData;
	WSAStartup(0x0202,&wsaData); /*windows socket startup */

	memset((char*)&serverAddr,0,sizeof(serverAddr));
	serverAddr.sin_family=AF_INET; /*set server address protocol family*/
	serverAddr.sin_addr.s_addr=INADDR_ANY;
	serverAddr.sin_port=htons((u_short)serverPort);  /*set port*/

	sock = socket(PF_INET,SOCK_DGRAM,0);/*create a socket*/
	if ( sock < 0 ) {
		Log("socket() error: %d", WSAGetLastError());
		return ;
	}

	if (bind(sock,(LPSOCKADDR)&serverAddr,len)<0) {
		Log("bind() error: %d", WSAGetLastError());
		return ;
	}

	int buf = 1000000;
	if ( setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char*) &buf, sizeof(buf)) == SOCKET_ERROR ) {
		Log("setsockopt() error: %d", WSAGetLastError());
		return ;
	}

	BOOL reuse = TRUE;
	if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char*) &reuse, sizeof(reuse)) == SOCKET_ERROR ) {
		Log("setsockopt() error: %d", WSAGetLastError());
		return ;
	}


	// Do not return socket error if remote host returns ICMP_PORT_UNREACHABLE
	DWORD dwBytesReturned = 0;
	BOOL bNewBehavior = FALSE;

#ifndef SIO_UDP_CONNRESET
    #define SIO_UDP_CONNRESET _WSAIOW(IOC_VENDOR,12)
#endif

	if ( WSAIoctl(sock, SIO_UDP_CONNRESET, 
		&bNewBehavior, sizeof(bNewBehavior), NULL, 0, &dwBytesReturned,
		NULL, NULL) == SOCKET_ERROR ) {
			Log("WSAIoctl() error: %d", WSAGetLastError());
			return ;
	}
	
	if (debug) Log("UDP server created");


	bInitialized = TRUE;


}

void OrbMaster::clbkSimulationStart(RenderMode mode)
{
	if ( ! bInitialized )
		Init();

	oapiSaveScenario("000MasterSlave", "");
	lastSaveT = oapiGetSysTime();
	lastUpdateT = oapiGetSysTime();

	for (int i=0; i<oapiGetVesselCount(); i++) {
		VESSEL *v = oapiGetVesselInterface(oapiGetVesselByIndex(i));
		if (debug) Log("Vessel #%d: name=%s class=%s", i, v->GetName(), v->GetClassNameA());
	}

	HookKeyboard("OrbMaster.dll");
}

void OrbMaster::clbkSimulationEnd(void)
{
	UnHookKeyboard();
}

void OrbMaster::clbkPreStep(double simt, double simdt, double mjd)
{
}

int OrbMaster::FindClient(struct sockaddr_in *client_in)
{
	for (int k=0; k<clients.size(); k++) {
		struct sockaddr_in *c_in = (sockaddr_in *) &clients[k];
		if ( ( client_in->sin_addr.S_un.S_addr == c_in->sin_addr.S_un.S_addr ) &&
				( client_in->sin_port == c_in ->sin_port ) ) {
					return k;
		}
	}

	return -1;
}

void OrbMaster::clbkPostStep(double simt, double simdt, double mjd)
{
	if ( ! bInitialized )
		return ;

	if ( oapiGetSysTime() - lastSaveT > 10 ) {
		oapiSaveScenario("000MasterSlave", "");
		lastSaveT = oapiGetSysTime();
	}

	if ( oapiGetSysTime() - lastUpdateT <= 1.0/maxUpdateRate )
		return ;
	lastUpdateT = oapiGetSysTime();

	StartMessage start;
	start.Send(sock, &clients);

	FocusMessage focus;
	focus.Send(sock, &clients);;

	KeyboardMessageBuf.Send(sock, &clients);
	KeyboardMessageBuf.Clear();

	for (int i=0; i<oapiGetVesselCount(); i++) {
		Message *vmsg = CreateVesselMessage(oapiGetVesselByIndex(i));
		if (vmsg) {
			vmsg->Send(sock, &clients);
			delete vmsg;
		}

		VesselExtraMessage ve(oapiGetVesselByIndex(i));
		ve.Send(sock, &clients);
	
		for ( int j=0; j<textSyncClasses.size(); j++ ) {
			char *cl = oapiGetVesselInterface(oapiGetVesselByIndex(i))->GetClassNameA();
			if ( stricmp(textSyncClasses.at(j), cl) == 0 ) {
				//Log("text sync for %s", oapiGetVesselInterface(oapiGetVesselByIndex(i))->GetName());
				Message *vsmsg = CreateVesselStateMessage(oapiGetVesselByIndex(i));
				if (vsmsg) {
					vsmsg->Send(sock, &clients);
					delete vsmsg;
				}
				break;
			}
					
		}

		AirfoilMessage amsg(oapiGetVesselByIndex(i));
		amsg.Send(sock, &clients);

		AnimationMessage anmsg(oapiGetVesselByIndex(i));
		anmsg.Send(sock, &clients);

		XRMessage xrmsg(oapiGetVesselByIndex(i));
		if ( xrmsg.IsValid() )
			xrmsg.Send(sock, &clients);
	}
	EndMessage end;
	end.Send(sock, &clients);
	DWORD sendTime = GetTickCount();

	vector<int> acks;
	for (int i=0; i<clients.size(); i++)
		acks.push_back(0);

	int maxWait = 30;
	for (int j=0; j<maxWait; j++) {
		while ( MessageSocketHasData(sock) ) {
			struct sockaddr clientAddr;
			int len = sizeof(clientAddr);

			Message *msg = ReceiveMessage(sock, (sockaddr*) &clientAddr, &len);
			if ( ! MESSAGE_VALID(msg) ) 
				break;

			struct sockaddr_in *client_in = (sockaddr_in *) &clientAddr;


			if ( msg->GetType() == MESSAGE_TYPE_HELLO ) {
				int k = FindClient(client_in);
				if ( k < 0 ) {
					// New client
					clients.push_back(clientAddr);
					clientLastAck.push_back(oapiGetSysTime());

					if (debug) Log("Client %s:%d subscribed", inet_ntoa(client_in->sin_addr), ntohs(client_in->sin_port));
				}
			}

			if ( msg->GetType() == MESSAGE_TYPE_ACK ) {
					int k = FindClient(client_in);
					if ( k>=0 ) {
						acks[k]++;
						clientLastAck[k] = oapiGetSysTime();
					}
			}

			if ( readOnly == 0 ) {
				if ( msg->GetType() == MESSAGE_TYPE_KEYBOARD ) {
					KeyboardMessage *kmsg = (KeyboardMessage *)msg;
					kmsg->SetState();
				}
			}

			delete msg;
		}

		if ( waitForClients && clients.size() == 0 )
			oapiSetPause(true);

		// Do not wait for acknowledgement, render and send next frame
		if ( ! waitForACK )
			break;

		int allAcked = 1;
		for (int k=0; k<acks.size(); k++) {
			if ( acks[k] == 0 )
				allAcked = 0;
		}
		if ( allAcked ) {
			//Log("All ACK'd");
			break;
		}


		Sleep(1);
	}

	
	char buf[256];
	if (acks.size() > 0) {
		strcpy(buf, "CLIENTS:");
		for (int i=0; i<acks.size(); i++) {
			struct sockaddr_in *c = (sockaddr_in*) &clients[i];

			if ( acks[i] ) {
				sprintf(buf+strlen(buf), " %15s:%5d", inet_ntoa(c->sin_addr), c->sin_port);
			} else {
				sprintf(buf+strlen(buf), " %15s:%5d", inet_ntoa(c->sin_addr), c->sin_port);
			}
		} 
	} else {
		strcpy(buf, "NO CLIENTS");
	}
	if (debug) strcpy(oapiDebugString(), buf);

	for (int k=0; k<clientLastAck.size(); k++ ) {
		if ( oapiGetSysTime() - clientLastAck[k] > 3 ) {
			// drop the client
			if (debug) Log("Client #%d timed out", k);
			clients.erase(clients.begin() + k);
			clientLastAck.erase(clientLastAck.begin() + k);
		}
	}
}

DLLCLBK void InitModule (HINSTANCE hDLL)
{
	initLog("OrbMaster.log");
	//Log("DLL loaded, build: %s %s", __DATE__, __TIME__);

	hDllInstance = hDLL; // remember the instance handle

	OrbMaster *mod = new OrbMaster(hDllInstance);
	oapiRegisterModule((oapi::Module*) mod);
}


DLLCLBK void ExitModule (HINSTANCE hDLL)
{
	closeLog();
}

