#include "stdafx.h"

#include "..\\EasyHook\\EasyHook.h"
#pragma comment(lib, "..\\EasyHook\\EasyHook32.lib")

#include "Orbitersdk.h"
#include "Log.h"

int enable_SetAnimation = 0;
int HookDebug = 0;

class VesselDetour
{
public:
	bool SetAnimation(UINT anim, double state)   const;
	static bool (VesselDetour::*Real_SetAnimation)(UINT anim, double state) const;
	static bool (VesselDetour::*Hook_SetAnimation)(UINT anim, double state) const;
};

bool (VesselDetour::* VesselDetour::Real_SetAnimation)(UINT anim, double state) const = 
	(bool (VesselDetour::*)(UINT anim, double state) const) &VESSEL::SetAnimation;

bool (VesselDetour::* VesselDetour::Hook_SetAnimation)(UINT anim, double state) const = 
	&VesselDetour::SetAnimation;

bool VesselDetour::SetAnimation(UINT anim, double state) const
{
	if ( enable_SetAnimation ) {
		//Log("VesselDetour::SetAnimation ACTIVE anim=%d state=%f", anim, state);
		return (this->*Real_SetAnimation)(anim, state);
	}
	else {
		//Log("VesselDetour::SetAnimation DUMMY anim=%d state=%f", anim, state);
		return true;
	}
}

TRACED_HOOK_HANDLE hHook;

void HookVessel(void) 
{
	if (HookDebug) Log("Hooking vessel...");
	if ( oapiGetOrbiterVersion() != 100830 ) {
		Log("ERROR: Unsupported orbiter version!");
		return ;
	}

	VesselDetour *vd = new VesselDetour();
	void **target = (void**)&vd->Real_SetAnimation;
	void **hook = (void**)&vd->Hook_SetAnimation;
	if (HookDebug) Log("VESSEL::SetAnimation=%p", &VESSEL::SetAnimation);
	if (HookDebug) Log("VesselDetour::SetAnimation=%p", &VesselDetour::SetAnimation);
	if (HookDebug) Log("vd=%p", vd);
	if (HookDebug) Log("target=%p", target);
	if (HookDebug) Log("*target=%p", *target);
	if (HookDebug) Log("*hook=%p", *hook);

	hHook = new HOOK_TRACE_INFO();
	NTSTATUS ret = LhInstallHook(*target, *hook, NULL, hHook);
	if ( ret != 0 ) { 
		Log("ERROR: could not hook vessel, error %08x", ret);
		return ;
	}
	
	// activate the hook for the current thread
	ULONG  ACLEntries[1] = {0};
	LhSetInclusiveACL(ACLEntries, 1, hHook);

	if (HookDebug) Log("Vessel hooked");
}

void EnableSetAnimation(void)
{
	//Log("EnableSetAnimation");
	enable_SetAnimation = 1;
}

void DisableSetAnimation(void)
{
	//Log("DisableSetAnimation");
	enable_SetAnimation = 0;
}

void UnhookVessel(void) 
{
	if (HookDebug) Log("Unhooking vessel...");
    LhUninstallHook(hHook);
    // now we can safely release the traced handle
    delete hHook;
    hHook = NULL;
    // even if the hook is removed, we need to wait for memory release
    LhWaitForPendingRemovals();
	if (HookDebug) Log("Vessel unhooked");
}


