

#include "stdafx.h"

#define STRICT
#define ORBITER_MODULE
#include "Orbitersdk.h"
#include "Message.h"
#include "HookVessel.h"

//#include "vld.h"

HINSTANCE hDllInstance;

class OrbSlave : oapi::Module {
private:
	OBJHANDLE ghost;
	VESSELSTATUS status;
	SOCKET sock;
	int firstSerial;
	int lastSerial;
	int lost;
	int dropped;
	int transactions;
	int lastBufferSize;
	double lastSimT;
	struct sockaddr_in serverAddr;
	int switchFocus;
	int debug;
	int waitForMaster;
	int sendKeys;
	double azimuth, elevation;
	double dt_rms;
	double dt_mean;
	double dt_mean_last;

public:
	OrbSlave(HINSTANCE hDLL);
	void ReadState(void);
	virtual void clbkSimulationStart(RenderMode mode);
	virtual void clbkPreStep(double simt, double simdt, double mjd);
	virtual void clbkPostStep(double simt, double simdt, double mjd);
	virtual void clbkSimulationEnd(void);
	void UpdateCounters(Message *msg);
	u_long SocketHasData(void);
};

void OrbSlave::UpdateCounters(Message *msg)
{
	if ( firstSerial == -1 ) {
		firstSerial = msg->GetSerial();
		lastSerial = msg->GetSerial();
	} else {
		lost += msg->GetSerial() - lastSerial - 1;
		lastSerial = msg->GetSerial();
	}
};

OrbSlave::OrbSlave(HINSTANCE hDLL) : oapi::Module(hDLL) 
{ 
	FILEHANDLE f = oapiOpenFile("OrbSlave.cfg", FILE_IN, CONFIG);
	if ( ! oapiReadItem_int(f, "debug", debug ) ) {
		debug = 0;
	}
	if (debug) Log("Config: debug=%d", debug);

	char serverIP[256];
	if ( ! oapiReadItem_string(f, "masterIP", serverIP) ) {
		strcpy(serverIP, "127.0.0.1");
	}
	if (debug) Log("Config: masterIP=%s", serverIP);

	int serverPort;
	if ( ! oapiReadItem_int(f, "masterPort", serverPort ) ) {
		serverPort = 40000;
	}
	if (debug) Log("Config: masterPort=%d", serverPort);

	if ( ! oapiReadItem_int(f, "switchFocus", switchFocus ) ) {
		switchFocus = 0;
	}
	if (debug) Log("Config: switchFocus=%d", switchFocus);

	if ( ! oapiReadItem_int(f, "waitForMaster", waitForMaster ) ) {
		waitForMaster = 0;
	}
	if (debug) Log("Config: waitForMaster=%d", waitForMaster);

	if ( ! oapiReadItem_int(f, "sendKeys", sendKeys ) ) {
		sendKeys = 0;
	}
	if (debug) Log("Config: sendKeys=%d", sendKeys);

	if ( ! oapiReadItem_float(f, "elevation", elevation ) ) {
		elevation = 0;
	}
	if (debug) Log("Config: elevation=%g", elevation);

	if ( ! oapiReadItem_float(f, "azimuth", azimuth ) ) {
		azimuth = 0;
	}
	if (debug) Log("Config: azimuth=%g", azimuth);

	WSADATA wsaData;
	WSAStartup(0x0202,&wsaData); /*windows socket startup */

	struct sockaddr_in clientAddr;
	int len = sizeof(clientAddr);
	clientAddr.sin_family = AF_INET;
	clientAddr.sin_addr.s_addr = INADDR_ANY;
	clientAddr.sin_port = htons(0);

	sock = socket(PF_INET,SOCK_DGRAM,0);/*create a socket*/
	if ( sock < 0 ) {
		Log("socket() error: %d", WSAGetLastError());
		return ;
	}

	if (bind(sock,(LPSOCKADDR)&clientAddr,len)<0) {
		Log("bind() error: %d", WSAGetLastError());
		return ;
	}

	int buf = 1000000;
	if ( setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char*) &buf, sizeof(buf)) == SOCKET_ERROR ) {
		Log("setsockopt() error: %d", WSAGetLastError());
		return ;
	}

	// Do not return socket error if remote host returns ICMP_PORT_UNREACHABLE
	DWORD dwBytesReturned = 0;
	BOOL bNewBehavior = FALSE;

#ifndef SIO_UDP_CONNRESET
    #define SIO_UDP_CONNRESET _WSAIOW(IOC_VENDOR,12)
#endif

	if ( WSAIoctl(sock, SIO_UDP_CONNRESET, 
		&bNewBehavior, sizeof(bNewBehavior), NULL, 0, &dwBytesReturned,
		NULL, NULL) == SOCKET_ERROR ) {
			Log("WSAIoctl() error: %d", WSAGetLastError());
			return ;
	}

/*	u_long nonblock = 1;
	if (ioctlsocket(sock, FIONBIO, &nonblock) != NO_ERROR) {
		Log("ioctlsocket() error: %d", WSAGetLastError());
		return ;
	}
	*/

	struct sockaddr_in myAddr;
	int len2 = sizeof(myAddr);
	getsockname(sock, (struct sockaddr*)&myAddr, &len2);
	if ( debug )
		Log("Bound to %s:%d", inet_ntoa(myAddr.sin_addr), myAddr.sin_port);

	serverAddr.sin_family = AF_INET;
	serverAddr.sin_addr.s_addr = inet_addr(serverIP);
	serverAddr.sin_port = htons((u_short)serverPort);

	firstSerial = -1;
	lost = 0;
	dropped = 0;
	lastBufferSize = buf;
	lastSimT = 0;
	transactions = 0;
	dt_mean_last = 0;
}

u_long OrbSlave::SocketHasData(void)
{
	return MessageSocketHasData(sock);
}

void OrbSlave::ReadState(void)
{

	Message *msg;

	VESSEL *v = oapiGetFocusInterface();

	BOOL bContinue = TRUE;
	do {
		if ( waitForMaster == 0 && ( ! SocketHasData() ) ) {
			//Log("no data from master!");
			return ;
		}


		msg = ReceiveMessage(sock, 100);
		if ( msg == MESSAGE_ERROR ) {
			Log("ReceiveMessage() error");
			return;
		}

		if ( msg == MESSAGE_TIMED_OUT ) {
			if ( debug ) 
				Log("ReceiveMessage() timed out");
			return;
		}

		UpdateCounters(msg);


		if ( msg->GetType() == MESSAGE_TYPE_START ) {
			double dt = 86400*(((StartMessage*)msg)->GetMJD() - oapiGetSimMJD());
			dt_rms = dt_rms + (dt-dt_mean_last)*(dt-dt_mean_last);
			dt_mean = dt_mean + dt;
			if ( abs(dt) > 0.1 ) {
				if ( debug )
					Log("MJD sync delta=%g s", dt);
				oapiSetSimMJD( ((StartMessage*)msg)->GetMJD(), PROP_ORBITAL_FIXEDSURF | PROP_SORBITAL_FIXEDSURF );
			}

			oapiSetTimeAcceleration( ((StartMessage*)msg)->GetTimeAcceleration() );
		}

		if ( msg->GetType() == MESSAGE_TYPE_END ) {
			transactions++;

			int bufferSize = SocketHasData();
			if ( bufferSize == 0 ) {
				// transaction completed
				bContinue = FALSE;
			} else {
				if ( bufferSize > lastBufferSize ) {
					// We're lagging behind the server. Drop everything!
					while ( SocketHasData() ) {
						Message *m = ReceiveMessage(sock);
						delete m;
						dropped++;
					}
				}
				lastBufferSize = bufferSize;
				bContinue = FALSE;
			}
		}

		if ( msg->GetType() == MESSAGE_TYPE_VESSEL ) {	
			VesselMessage *vmsg = (VesselMessage *)msg;
			vmsg->UpdateVessel();
		}

		if ( msg->GetType() == MESSAGE_TYPE_VESSEL2 ) {	
			Vessel2Message *vmsg = (Vessel2Message *)msg;
			vmsg->UpdateVessel();
		}

		if ( msg->GetType() == MESSAGE_TYPE_VESSELSTATE ) {	
			VesselStateMessage *vsmsg = (VesselStateMessage *)msg;
			vsmsg->UpdateVessel();
		}

		if ( msg->GetType() == MESSAGE_TYPE_AIRFOIL ) {
			AirfoilMessage *amsg = (AirfoilMessage *)msg;
			amsg->UpdateVessel();
		}

		if ( msg->GetType() == MESSAGE_TYPE_XR ) {
			XRMessage *xrmsg = (XRMessage *)msg;
			xrmsg->UpdateVessel();
		}

		if ( msg->GetType() == MESSAGE_TYPE_FOCUS ) {
			if ( switchFocus ) {
				FocusMessage *fmsg = (FocusMessage*) msg;
				fmsg->SwitchFocus();
			}
		}

		if ( msg->GetType() == MESSAGE_TYPE_KEYBOARD ) {
			KeyboardMessage *kmsg = (KeyboardMessage *)msg;
			kmsg->SetState();
		}

		if ( msg->GetType() == MESSAGE_TYPE_VESSELEXTRA ) {
			VesselExtraMessage *vemsg = (VesselExtraMessage *)msg;
			vemsg->UpdateVessel();
		}
		
		if ( msg->GetType() == MESSAGE_TYPE_ANIMATION ) {
			AnimationMessage *amsg = (AnimationMessage *)msg;
			amsg->UpdateVessel();
		}

		delete msg;
	} while (bContinue);

	AckMessage ack;
	ack.Send(sock, (struct sockaddr*) &serverAddr);	
}

void OrbSlave::clbkSimulationStart(RenderMode mode)
{
	HookDebug = debug;
	HookVessel();

	HelloMessage hello;
	hello.Send(sock, (sockaddr*) &serverAddr);

	if ( sendKeys )
		HookKeyboard("OrbSlave.dll");

	oapiCameraSetCockpitDir(azimuth*PI/180, elevation*PI/180);
}

void OrbSlave::clbkSimulationEnd(void)
{
	UnhookVessel();

	if ( sendKeys )
		UnHookKeyboard();
}

void OrbSlave::clbkPreStep(double simt, double simdt, double mjd)
{
	//Log("clbkPreStep");
}

void OrbSlave::clbkPostStep(double simt, double simdt, double mjd)
{
	if ( sendKeys ) {
		KeyboardMessageBuf.Send(sock, (struct sockaddr*) &serverAddr);
		KeyboardMessageBuf.Clear();
	}

	EnableSetAnimation();
	ReadState();
	DisableSetAnimation();

	if ( oapiGetSysTime() - lastSimT > 1 ) {
		HelloMessage hello;
		hello.Send(sock, (sockaddr*) &serverAddr);

		if ( debug ) 
			Log("Pkts: %d lost: %d (%4.1f %%) dropped: %d (%4.1f %%) %d upd/sec mean_delay=%4.1f ms jitter=%4.1f ms rms",
				(lastSerial-firstSerial), 
				lost, 100.0*lost/(lastSerial-firstSerial), 
				dropped, 100.0*dropped/(lastSerial-firstSerial),
				transactions,
				1000.0*dt_mean/transactions,
				1000.0*sqrt(dt_rms)/(transactions-1)
				);
				

		firstSerial = -1;
		lost = 0;
		dropped = 0;
		transactions = 0;
		dt_rms = 0;
		dt_mean = 0;

		lastSimT = oapiGetSysTime();
	}
}
	

DLLCLBK void InitModule (HINSTANCE hDLL)
{
	initLog("OrbSlave.log");
	//Log("DLL loaded, build: %s %s", __DATE__, __TIME__);

	hDllInstance = hDLL; // remember the instance handle

	OrbSlave *mod = new OrbSlave(hDllInstance);
	oapiRegisterModule((oapi::Module*) mod);
}


DLLCLBK void ExitModule (HINSTANCE hDLL)
{
	closeLog();
}

